#include "global.h"
#include "hid.h"
#include "out.h"
#include "kb.h"
#include <stdio.h>
#include <syslog.h>
#include <stdlib.h>
#include <linux/input.h>


/* Bits in Input/Output/Feature reports */
#define	RPT_CONST	0x001	/* else it's data */
#define	RPT_VAR		0x002	/* else it's a bit set */
#define	RPT_REL		0x004	/* else it's absolute */
#define	RPT_WRAP	0x008
#define	RPT_NONLIN 	0x010
#define	RPT_NOPREF 	0x020
#define	RPT_NULL	0x040
#define	RPT_VOL		0x080
#define	RPT_BUFF	0x100	/* else it's a bit field */

struct hid_report_field {
	unsigned long usage;
	unsigned flags;
	long	min, max;
	int	bits;
};

struct hid_report {
	unsigned id;
	int	fields;
	struct hid_report_field *field;
};

struct hid_report *hid_report;
int	hid_report_count;


static
void
hid_report_descriptor_dump (unsigned char *data, int bytes) {

	int	i, j;
	int	tag, type, size;
	unsigned long val;

	printf ("HID report descriptor (%d bytes)\n", bytes);

	for (i = 0; i < bytes; ) {

		tag  = data [i] >> 4 & 0xf;
		type = data [i] >> 2 & 0x3;
		size = data [i] >> 0 & 0x3;
		i++;

		if (tag == 0xf) {
			/* long item */
			size = data [i++];
			tag = data [i++];
		} else
			size = 1 << size >> 1;

		for (val = j = 0; j < size; j++)
			val |= data [i++] << (j << 3);

		switch (type) {

		case 0:	/* Main */
			switch (tag) {
			case 0x8: printf ("M Input 0x%lx\n", val); break;
			case 0x9: printf ("M Output 0x%lx\n", val); break;
			case 0xa: printf ("M Collect 0x%lx\n", val); break;
			case 0xb: printf ("M Feature 0x%lx\n", val); break;
			case 0xc: printf ("M EndColl 0x%lx\n", val); break;
			default:  printf ("M UNKNOWN(0x%x) 0x%lx\n", tag, val);
			}
			break;

		case 1:	/* Global */
			switch (tag) {
			case 0x0: printf ("G Usage Page 0x%lx\n", val); break;
			case 0x1:
				if (size < 4)
					val |= -(val >> (size << 3)-1) <<
						(size << 3);
				printf ("G Log Min 0x%lx\n", val);
				break;
			case 0x2:
				if (size < 4)
					val |= -(val >> (size << 3)-1) <<
						(size << 3);
				printf ("G Log Max 0x%lx\n", val);
				break;
			case 0x7: printf ("G Rpt Sz 0x%lx\n", val); break;
			case 0x8: printf ("G Rpt Id 0x%lx\n", val); break;
			case 0x9: printf ("G Rpt Ct 0x%lx\n", val); break;
			case 0xa: printf ("G Push 0x%lx\n", val); break;
			case 0xb: printf ("G Pop 0x%lx\n", val); break;
			default:  printf ("G UNKNOWN(0x%x) 0x%lx\n", tag, val);
			}
			break;

		case 2: /* Local */
			switch (tag) {
			case 0x0: printf ("L Usage 0x%lx\n", val); break;
			case 0x1: printf ("L Usage Min 0x%lx\n", val); break;
			case 0x2: printf ("L Usage Max 0x%lx\n", val); break;
			default:  printf ("L UNKNOWN(0x%x) 0x%lx\n", tag, val);
			}
			break;

		default: printf ("UNKNOWN TYPE(%d) tag %d 0x%lx\n",
				 type, tag, val);
		}
	}
}


/*
 *	Parse the report definitions from the HID
 *	descriptor obtained as an SDP attribute into
 *	a list of Input reports that the device may
 *	emit and the definitions of the fields therein.
 */

static
void
add_field_to_report (unsigned id,
		     unsigned long usage, unsigned flags,
		     long min, long max, int bits) {

	int	i;
	struct hid_report_field *f;

	for (i = 0; i < hid_report_count; i++)
		if (hid_report [i].id == id)
			break;
	if (i == hid_report_count) {
		hid_report = realloc (hid_report,
				sizeof *hid_report * (hid_report_count+1));
		memset (&hid_report [hid_report_count], 0,
			sizeof *hid_report);
		hid_report [hid_report_count++].id = id;
	}

	hid_report [i].field = realloc (hid_report [i].field,
					sizeof *hid_report [i].field *
					  (hid_report [i].fields + 1));
	f = &hid_report [i].field [hid_report [i].fields++];
	f->usage = usage;
	f->flags = flags;
	f->min = min;
	f->max = max;
	f->bits = bits;
}

struct hid_global {
	unsigned usage_page,
		logical_min,
		logical_max,
		report_size,	/* in bits */
		report_id,
		report_count;
};

struct hid_local {
	unsigned long usage_min, usage_max;
};

#define	MAX_GSTACK	10
#define	MAX_LOCAL	64


void
hid_report_descriptor (unsigned char *data, int bytes) {

	int	i, j, k;
	int	tag, type, size;
	int	gstk_top = 0;
	int	local_ct = 0;
	unsigned long val, usage;
	struct hid_global gstk [MAX_GSTACK];
	struct hid_global gst;
	struct hid_local loc [MAX_LOCAL];

	if (verbosity > 1)
		hid_report_descriptor_dump (data, bytes);

	memset (&gstk [0], 0, sizeof gstk [0]);
	memset (&gst, 0, sizeof gst);

	for (i = 0; i < bytes; ) {

		tag  = data [i] >> 4 & 0xf;
		type = data [i] >> 2 & 0x3;
		size = data [i] >> 0 & 0x3;
		i++;

		if (tag == 0xf) {
			/* long item */
			size = data [i++];
			tag = data [i++];
		} else
			size = 1 << size >> 1;

		for (val = j = 0; j < size; j++)
			val |= data [i++] << (j << 3);

		switch (type) {

		case 0:	/* Main */
			switch (tag) {
			case 0x8:	/* Input */
				k = 0;
				usage = local_ct ? loc [k].usage_min : 0;
				for (j = 0; j < gst.report_count; j++) {
					add_field_to_report (gst.report_id,
							     usage, val,
							     gst.logical_min,
							     gst.logical_max,
							     gst.report_size);
					if (local_ct &&
					    (val & RPT_VAR &&
					     usage++ == loc [k].usage_max ||
					     !(val & RPT_VAR))) {
						if (k < local_ct-1)
							k++;
						usage = loc [k].usage_min;
					}
				}
				break;
			case 0x9:	/* Output */
				break;
			case 0xa:	/* Collection */
				switch (val) {
				case 0:	/* Physical */
					break;
				case 1:	/* Application */
					break;
				case 2:	/* Logical */
					break;
				case 3:	/* Report */
					break;
				case 4:	/* Named Array */
					break;
				case 5:	/* Usage Switch */
					break;
				case 6:	/* Usage Modifier */
					break;
				}
				break;
			case 0xb:	/* Feature */
				break;
			case 0xc:	/* EndCollection */
				break;
			}
			local_ct = 0;
			break;

		case 1:	/* Global */
			switch (tag) {
			case 0x0:
				gst.usage_page = val;
				break;
			case 0x1:
				if (size < 4)
					val |= -(val >> (size << 3)-1) <<
						(size << 3);
				gst.logical_min = val;
				break;
			case 0x2:
				if (size < 4)
					val |= -(val >> (size << 3)-1) <<
						(size << 3);
				gst.logical_max = val;
				break;
			case 0x7:
				gst.report_size = val;
				break;
			case 0x8:
				gst.report_id = val;
				break;
			case 0x9:
				gst.report_count = val;
				break;
			case 0xa:	/* Push */
				if (gstk_top < MAX_GSTACK)
					gstk [gstk_top++] = gst;
				break;
			case 0xb:	/* Pop */
				if (gstk_top > 0)
					gst = gstk [--gstk_top];
				break;
			}
			break;

		case 2: /* Local */
			switch (tag) {
			case 0x0:	/* Usage */
			case 0x1:	/* Usage Minimum */
				if (size <= 2)
					val |= gst.usage_page << 16;
				if (local_ct < MAX_LOCAL) {
					memset (&loc [local_ct], 0,
						sizeof loc [0]);
					loc [local_ct].usage_min = val;
					loc [local_ct++].usage_max = val;
				}
				break;
			case 0x2:	/* Usage Maximum */
				if (size <= 2)
					val |= gst.usage_page << 16;
				if (local_ct)
					loc [local_ct-1].usage_max = val;
				break;
			}
			break;
		}
	}

	if (i > bytes)
		syslog (LOG_WARNING, "HID parser overran data buffer "
			"(index %d > %d bytes)", i, bytes);

	if (verbosity > 1) {
		printf ("HID parsing done; %d reports\n", hid_report_count);
		for (i = 0; i < hid_report_count; i++) {
			printf ("report ID 0x%x: %d fields\n",
				hid_report [i].id, hid_report [i].fields);
			for (j = 0; j < hid_report [i].fields; j++)
				printf ("%d bits: usage 0x%lx flags 0x%x "
					"[%ld-%ld]\n",
					hid_report [i].field [j].bits,
					hid_report [i].field [j].usage,
					hid_report [i].field [j].flags,
					hid_report [i].field [j].min,
					hid_report [i].field [j].max);
		}
	}
}


/*
 *	Mouse state
 */

static	unsigned mouse_butt;		/* mouse button mask */

void
mouse_reset (void) {

	int	i;

	for (i = BTN_MOUSE; mouse_butt; i++) {
		if (mouse_butt & 1)
			out_event (EV_KEY, i, 0);
		mouse_butt >>= 0;
	}
}


/*
 *	Process a HID input report received as a DATA packet
 *	on its Interrupt channel
 */

static	int	key_down [32];		/* keyboard set */
static	int	keys_down;
static	int	any_keyboard;


static	int	hid_to_linux [256] = {

/* 0x00 */	0 /* no event */, 0 /* rollover error */, 0 /* POST fail */,
		0 /* undefined error */, KEY_A, KEY_B, KEY_C, KEY_D,
/* 0x08 */	KEY_E, KEY_F, KEY_G, KEY_H, KEY_I, KEY_J, KEY_K, KEY_L,
/* 0x10 */	KEY_M, KEY_N, KEY_O, KEY_P, KEY_Q, KEY_R, KEY_S, KEY_T,
/* 0x18 */	KEY_U, KEY_V, KEY_W, KEY_X, KEY_Y, KEY_Z, KEY_1, KEY_2,
/* 0x20 */	KEY_3, KEY_4, KEY_5, KEY_6, KEY_7, KEY_8, KEY_9, KEY_0,
/* 0x28 */	KEY_ENTER, KEY_ESC, KEY_BACKSPACE, KEY_TAB,
		KEY_SPACE, KEY_MINUS, KEY_EQUAL, KEY_LEFTBRACE,
/* 0x30 */	KEY_RIGHTBRACE, KEY_BACKSLASH, KEY_BACKSLASH, KEY_SEMICOLON,
		KEY_APOSTROPHE, KEY_GRAVE, KEY_COMMA, KEY_DOT,
/* 0x38 */	KEY_SLASH, KEY_CAPSLOCK, KEY_F1, KEY_F2,
		KEY_F3, KEY_F4, KEY_F5, KEY_F6,
/* 0x40 */	KEY_F7, KEY_F8, KEY_F9, KEY_F10,
		KEY_F11, KEY_F12, KEY_SYSRQ, KEY_SCROLLLOCK,
/* 0x48 */	KEY_PAUSE, KEY_INSERT, KEY_HOME, KEY_PAGEUP,
		KEY_DELETE, KEY_END, KEY_PAGEDOWN, KEY_RIGHT,
/* 0x50 */	KEY_LEFT, KEY_DOWN, KEY_UP, KEY_NUMLOCK,
		KEY_KPSLASH, KEY_KPASTERISK, KEY_KPMINUS, KEY_KPPLUS,
/* 0x58 */	KEY_KPENTER, KEY_KP1, KEY_KP2, KEY_KP3,
		KEY_KP4, KEY_KP5, KEY_KP6, KEY_KP7,
/* 0x60 */	KEY_KP8, KEY_KP9, KEY_KP0, KEY_KPDOT, KEY_102ND, 0, 0, 0,
/* pmk!! fill these in, even though my keyboard doesn't use them... */
/* 0x68 */	0, 0, 0, 0, 0, 0, 0, 0,
/* 0x70 */	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
/* 0x80 */	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
/* 0x90 */	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
/* 0xa0 */	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
/* 0xb0 */	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
/* 0xc0 */	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
/* 0xd0 */	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
/* 0xe0 */	KEY_LEFTCTRL, KEY_LEFTSHIFT, KEY_LEFTALT,
		KEY_LEFTMETA,	/* Would COMPOSE be better? */
		KEY_RIGHTCTRL, KEY_RIGHTSHIFT, KEY_RIGHTALT, 0,
/* 0xe8 */	0, 0, 0, 0, 0, 0, 0, 0,
/* 0xf0 */	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0
};


static	int	hid_dvorak_to_linux [256] = {

/* 0x00 */	0 /* no event */, 0 /* rollover error */, 0 /* POST fail */,
		0 /* undefined error */, KEY_A, KEY_X, KEY_J, KEY_E,
/* 0x08 */	KEY_DOT, KEY_U, KEY_I, KEY_D, KEY_C, KEY_H, KEY_T, KEY_N,
/* 0x10 */	KEY_M, KEY_B, KEY_R, KEY_L, KEY_APOSTROPHE, KEY_P, KEY_O, KEY_Y,
/* 0x18 */	KEY_G, KEY_K, KEY_COMMA, KEY_Q, KEY_F, KEY_SEMICOLON,
			KEY_1, KEY_2,
/* 0x20 */	KEY_3, KEY_4, KEY_5, KEY_6, KEY_7, KEY_8, KEY_9, KEY_0,
/* 0x28 */	KEY_ENTER, KEY_ESC, KEY_BACKSPACE, KEY_TAB,
		KEY_SPACE, KEY_LEFTBRACE, KEY_RIGHTBRACE, KEY_SLASH,
/* 0x30 */	KEY_EQUAL, KEY_BACKSLASH, KEY_BACKSLASH, KEY_R,
		KEY_MINUS, KEY_GRAVE, KEY_W, KEY_V,
/* 0x38 */	KEY_Z, KEY_CAPSLOCK, KEY_F1, KEY_F2,
		KEY_F3, KEY_F4, KEY_F5, KEY_F6,
/* 0x40 */	KEY_F7, KEY_F8, KEY_F9, KEY_F10,
		KEY_F11, KEY_F12, KEY_SYSRQ, KEY_SCROLLLOCK,
/* 0x48 */	KEY_PAUSE, KEY_INSERT, KEY_HOME, KEY_PAGEUP,
		KEY_DELETE, KEY_END, KEY_PAGEDOWN, KEY_RIGHT,
/* 0x50 */	KEY_LEFT, KEY_DOWN, KEY_UP, KEY_NUMLOCK,
		KEY_KPSLASH, KEY_KPASTERISK, KEY_KPMINUS, KEY_KPPLUS,
/* 0x58 */	KEY_KPENTER, KEY_KP1, KEY_KP2, KEY_KP3,
		KEY_KP4, KEY_KP5, KEY_KP6, KEY_KP7,
/* 0x60 */	KEY_KP8, KEY_KP9, KEY_KP0, KEY_KPDOT, KEY_102ND, 0, 0, 0,
/* pmk!! fill these in, even though my keyboard doesn't use them... */
/* 0x68 */	0, 0, 0, 0, 0, 0, 0, 0,
/* 0x70 */	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
/* 0x80 */	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
/* 0x90 */	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
/* 0xa0 */	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
/* 0xb0 */	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
/* 0xc0 */	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
/* 0xd0 */	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
/* 0xe0 */	KEY_LEFTCTRL, KEY_LEFTSHIFT, KEY_LEFTALT,
		KEY_LEFTMETA,	/* Would COMPOSE be better? */
		KEY_RIGHTCTRL, KEY_RIGHTSHIFT, KEY_RIGHTALT, 0,
/* 0xe8 */	0, 0, 0, 0, 0, 0, 0, 0,
/* 0xf0 */	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0
};


static
void
hid_usage (unsigned long usage, unsigned flags, long val) {

	int	butt;
	unsigned buttmask;

	switch (usage >> 16) {

	case 0x0000:
		/* no usage */
		return;

	case 0x0001:	/* Generic Desktop */

		if (!val)
			return;

		switch (usage & 0xffff) {

		case 0x0030:	/* Generic desktop, X motion */
			if (flags & RPT_REL)
				out_event (EV_REL, REL_X, val);
			else
				out_event (EV_ABS, ABS_X, val);
			return;

		case 0x0031:	/* Generic desktop, Y motion */
			if (flags & RPT_REL)
				out_event (EV_REL, REL_Y, val);
			else
				out_event (EV_ABS, ABS_Y, val);
			return;

		case 0x0032:	/* Generic desktop, Z motion */
			if (flags & RPT_REL)
				out_event (EV_REL, REL_Z, val);
			else
				out_event (EV_ABS, ABS_Z, val);
			return;

		case 0x0036:	/* Generic desktop, slider motion */
			if (flags & RPT_REL) {
				out_event (EV_REL, REL_HWHEEL, val);
				return;
			}
			break;

		case 0x0037:	/* Generic desktop, dial motion */
			if (flags & RPT_REL) {
				out_event (EV_REL, REL_DIAL, val);
				return;
			}
			break;

		case 0x0038:	/* Generic desktop, wheel motion */
			if (flags & RPT_REL) {
				out_event (EV_REL, REL_WHEEL, val);
				return;
			}
			break;

		case 0x0082:	/* Generic desktop, sleep button */
			if (val == 1) {
				syslog (LOG_INFO, "Sleep button pressed "
					"(ignored)");
				return;
			}
			break;
		}
		break;

	case 0x0007:	/* Keyboard */
		any_keyboard = 1;
		if (val && usage & 0xffff) {
			if (dvorak_mode < 2)
				butt = hid_to_linux [usage & 0xffff];
			else
				butt = hid_dvorak_to_linux [usage & 0xffff];
			if (butt)
				key_down [keys_down++] = butt;
		}
		return;

	case 0x0009:	/* Buttons */
		switch (usage & 0xffff) {
		default:
		case 0:	return;
		case 1:	butt = BTN_LEFT;	break;
		case 2:	butt = BTN_RIGHT;	break;
		case 3:	if (swap_buttons_2_5)	return;
			butt = BTN_MIDDLE;	break;
		case 4:	butt = BTN_BACK;	break;
		case 5: butt = BTN_FORWARD; break;
		case 6:	butt = swap_buttons_2_5 ? BTN_MIDDLE : BTN_EXTRA; break;
		}
		buttmask = 1 << butt - BTN_MOUSE;
		if (val && !(mouse_butt & buttmask)) {
			out_event (EV_KEY, butt, 1);
			mouse_butt |= buttmask;
		} else if (!val && mouse_butt & buttmask) {
			out_event (EV_KEY, butt, 0);
			mouse_butt &= ~buttmask;
		}
		return;

	case 0x000c:	/* Consumer */
		if (!val)
			return;
		butt = 0;
		switch (usage & 0xffff) {
		case 0x95:	butt = KEY_HELP;	break;
		case 0xb5:	butt = KEY_NEXTSONG;	break;
		case 0xb6:	butt = KEY_PREVIOUSSONG;	break;
		case 0xb7:	butt = KEY_STOPCD;	break;
		case 0xcd:	butt = KEY_PLAYPAUSE;	break;
		case 0xe2:	butt = KEY_MUTE;	break;
		case 0xe9:	butt = KEY_VOLUMEUP;	break;
		case 0xea:	butt = KEY_VOLUMEDOWN;	break;
		case 0x183:	butt = KEY_CLOSECD;	break;	/* Media */
		case 0x18a:	butt = KEY_MAIL;	break;
		case 0x192:	butt = KEY_CALC;	break;
		case 0x199:	butt = KEY_COFFEE;	break;	/* Messenger */
		case 0x19c:	butt = KEY_SUSPEND;	break;	/* LogOff */
		case 0x1a7:	butt = KEY_COMPUTER;	break;	/* MyDocs */
		case 0x1ab:	butt = KEY_FILE;	break;	/* Save */
		case 0x1b6:	butt = KEY_PROG1;	break;	/* MyPicts */
		case 0x1b7:	butt = KEY_PROG2;	break;	/* MyMusic */
		case 0x201:	butt = KEY_FRONT;	break;	/* New */
		case 0x202:	butt = KEY_OPEN;	break;
		case 0x203:	butt = KEY_CLOSE;	break;
		case 0x207:	butt = KEY_PROG3;	break;	/* Spell */
		case 0x208:	butt = KEY_PROG4;	break;	/* Print */
		case 0x21a:	butt = KEY_UNDO;	break;
		case 0x223:	butt = KEY_HOMEPAGE;	break;
		case 0x279:	butt = KEY_AGAIN;	break;	/* Redo */
		case 0x28c:	butt = KEY_SENDFILE;	break;
		default:
			syslog (LOG_NOTICE, "unknown consumer usage 0x%lx",
				usage);
		case 0x289:	/* Reply */
		case 0x28b:	/* Forward */
			break;
		}
		if (butt) {
			out_event (EV_KEY, butt, 1);
			out_event (EV_KEY, butt, 0);
		}
		return;

	case 0xff00:	/* vendor specific */
		if (verbosity)
			syslog (LOG_INFO, "vendor HID usage 0x%lx value 0x%lx",
				usage, val);
		switch (usage & 0xffff) {
		case 0xfe01:
			if (is_battery_powered) {
				if (val == 1) {
					syslog (LOG_INFO, "adequate battery "
						"on %s",
						devname ? devname : bdaddr_str);
					low_battery = 0;
					return;
				} else if (val == 2) {
					syslog (LOG_NOTICE, "low battery on %s",
						devname ? devname : bdaddr_str);
					low_battery = 1;
					return;
				} else if (val == 3) {
					syslog (LOG_WARNING, "very low battery "
						"on %s",
						devname ? devname : bdaddr_str);
					low_battery = 1;
					return;
				}
			}
			break;
		case 0xfe03:
			if (val >= 0 && val <= 1)
				return;	/* FLock status change indicator */
			break;
		case 0xfe04:
			if (val == 1)
				return;	/* always follows FLock change */
			break;
		case 0xff02:
			if (!val)
				return;	/* consumer button press/release */
			break;
		}
		break;
	}

	syslog (LOG_NOTICE, "unknown HID usage (0x%lx + %ld)", usage, val);
	if (verbosity)
		printf ("unknown HID usage (0x%08lx + %ld)\n", usage, val);
}


void
hid_input_report (unsigned char *data, int bytes) {

	int	i, j;
	int	at = 0;
	long	val;
	unsigned id;
	unsigned long bits, bit;
	struct hid_report *r = NULL;
	struct hid_report_field *f;

	if (bytes <= 0)
		return;
	id = data [at >> 3], at += 8;

	if (!hid_report_count) {
		if (verbosity)
			printf ("input report received before HID "
				"descriptor has been processed\n");
		return;
	}

	for (i = 0; i < hid_report_count; i++) {
		r = &hid_report [i];
		if (id == r->id)
			break;
	}
	if (i == hid_report_count) {
		syslog (LOG_WARNING, "unexpected HID Input report 0x%x", id);
		return;
	}

	keys_down = 0;
	any_keyboard = 0;

	for (i = 0; i < r->fields; i++) {

		f = &r->field [i];

		bits = 0;
		for (j = 0; j < f->bits; j++) {
			bit = data [at >> 3] >> (at & 7) & 1;
			at++;
			if (!(f->flags & RPT_CONST))
				bits |= bit << j;
		}

		if (f->flags & RPT_VAR) {
			/* variable entry */
			val = bits;
			if (f->min < 0)
				val |= -(val >> f->bits-1 & 1) << f->bits;
			hid_usage (f->usage, f->flags, val);
		} else {
			/* array entry */
			hid_usage (f->usage + bits, f->flags, 1);
		}
	}

	if (any_keyboard)
		keys_now_down (key_down, keys_down);

	out_flush ();
}
